package org.apache.bcel.util;

import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.classfile.Visitor;
import org.apache.commons.lang3.StringUtils;
import org.apache.bcel.classfile.ConstantValue;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.classfile.Field;
import java.io.Writer;
import java.io.OutputStreamWriter;
import java.nio.charset.StandardCharsets;
import org.apache.bcel.classfile.Utility;
import org.apache.bcel.generic.ArrayType;
import java.util.Locale;
import org.apache.bcel.Const;
import org.apache.bcel.generic.Type;
import java.io.OutputStream;
import java.io.IOException;
import org.apache.bcel.classfile.ClassParser;
import org.apache.bcel.Repository;
import org.apache.bcel.generic.ConstantPoolGen;
import java.io.PrintWriter;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.EmptyVisitor;

public class BCELifier extends EmptyVisitor
{
    private static final String BASE_PACKAGE;
    private static final String CONSTANT_PREFIX;
    private final JavaClass clazz;
    private final PrintWriter printWriter;
    private final ConstantPoolGen constantPoolGen;
    
    static JavaClass getJavaClass(final String name) throws ClassNotFoundException, IOException {
        JavaClass javaClass;
        if ((javaClass = Repository.lookupClass(name)) == null) {
            javaClass = new ClassParser(name).parse();
        }
        return javaClass;
    }
    
    public static void main(final String[] argv) throws Exception {
        if (argv.length != 1) {
            System.out.println("Usage: BCELifier className");
            System.out.println("\tThe class must exist on the classpath");
            return;
        }
        final BCELifier bcelifier = new BCELifier(getJavaClass(argv[0]), System.out);
        bcelifier.start();
    }
    
    static String printArgumentTypes(final Type[] argTypes) {
        if (argTypes.length == 0) {
            return "Type.NO_ARGS";
        }
        final StringBuilder args = new StringBuilder();
        for (int i = 0; i < argTypes.length; ++i) {
            args.append(printType(argTypes[i]));
            if (i < argTypes.length - 1) {
                args.append(", ");
            }
        }
        return "new Type[] { " + args.toString() + " }";
    }
    
    static String printFlags(final int flags) {
        return printFlags(flags, FLAGS.UNKNOWN);
    }
    
    public static String printFlags(final int flags, final FLAGS location) {
        if (flags == 0) {
            return "0";
        }
        final StringBuilder buf = new StringBuilder();
        for (int i = 0, pow = 1; pow <= 32768; pow <<= 1, ++i) {
            if ((flags & pow) != 0x0) {
                if (pow == 32 && location == FLAGS.CLASS) {
                    buf.append(BCELifier.CONSTANT_PREFIX).append("ACC_SUPER | ");
                }
                else if (pow == 64 && location == FLAGS.METHOD) {
                    buf.append(BCELifier.CONSTANT_PREFIX).append("ACC_BRIDGE | ");
                }
                else if (pow == 128 && location == FLAGS.METHOD) {
                    buf.append(BCELifier.CONSTANT_PREFIX).append("ACC_VARARGS | ");
                }
                else if (i < Const.ACCESS_NAMES_LENGTH) {
                    buf.append(BCELifier.CONSTANT_PREFIX).append("ACC_").append(Const.getAccessName(i).toUpperCase(Locale.ENGLISH)).append(" | ");
                }
                else {
                    buf.append(String.format(BCELifier.CONSTANT_PREFIX + "ACC_BIT %x | ", pow));
                }
            }
        }
        final String str = buf.toString();
        return str.substring(0, str.length() - 3);
    }
    
    static String printType(final String signature) {
        final Type type = Type.getType(signature);
        final byte t = type.getType();
        if (t <= 12) {
            return "Type." + Const.getTypeName(t).toUpperCase(Locale.ENGLISH);
        }
        if (type.toString().equals("java.lang.String")) {
            return "Type.STRING";
        }
        if (type.toString().equals("java.lang.Object")) {
            return "Type.OBJECT";
        }
        if (type.toString().equals("java.lang.StringBuffer")) {
            return "Type.STRINGBUFFER";
        }
        if (type instanceof ArrayType) {
            final ArrayType at = (ArrayType)type;
            return "new ArrayType(" + printType(at.getBasicType()) + ", " + at.getDimensions() + ")";
        }
        return "new ObjectType(\"" + Utility.signatureToString(signature, false) + "\")";
    }
    
    static String printType(final Type type) {
        return printType(type.getSignature());
    }
    
    public BCELifier(final JavaClass clazz, final OutputStream out) {
        this.clazz = clazz;
        this.printWriter = new PrintWriter(new OutputStreamWriter(out, StandardCharsets.UTF_8), false);
        this.constantPoolGen = new ConstantPoolGen(this.clazz.getConstantPool());
    }
    
    private void printCreate() {
        this.printWriter.println("  public void create(OutputStream out) throws IOException {");
        final Field[] fields = this.clazz.getFields();
        if (fields.length > 0) {
            this.printWriter.println("    createFields();");
        }
        final Method[] methods = this.clazz.getMethods();
        for (int i = 0; i < methods.length; ++i) {
            this.printWriter.println("    createMethod_" + i + "();");
        }
        this.printWriter.println("    _cg.getJavaClass().dump(out);");
        this.printWriter.println("  }");
        this.printWriter.println();
    }
    
    private void printMain() {
        final String className = this.clazz.getClassName();
        this.printWriter.println("  public static void main(String[] args) throws Exception {");
        this.printWriter.println("    " + className + "Creator creator = new " + className + "Creator();");
        this.printWriter.println("    creator.create(new FileOutputStream(\"" + className + ".class\"));");
        this.printWriter.println("  }");
    }
    
    public void start() {
        this.visitJavaClass(this.clazz);
        this.printWriter.flush();
    }
    
    @Override
    public void visitField(final Field field) {
        this.printWriter.println();
        this.printWriter.println("    field = new FieldGen(" + printFlags(field.getAccessFlags()) + ", " + printType(field.getSignature()) + ", \"" + field.getName() + "\", _cp);");
        final ConstantValue cv = field.getConstantValue();
        if (cv != null) {
            this.printWriter.println("    field.setInitValue(" + cv + ")");
        }
        this.printWriter.println("    _cg.addField(field.getField());");
    }
    
    @Override
    public void visitJavaClass(final JavaClass clazz) {
        String className = clazz.getClassName();
        final String superName = clazz.getSuperclassName();
        final String packageName = clazz.getPackageName();
        final String inter = Utility.printArray(clazz.getInterfaceNames(), false, true);
        if (StringUtils.isNotEmpty((CharSequence)inter)) {
            className = className.substring(packageName.length() + 1);
            this.printWriter.println("package " + packageName + ";");
            this.printWriter.println();
        }
        this.printWriter.println("import " + BCELifier.BASE_PACKAGE + ".generic.*;");
        this.printWriter.println("import " + BCELifier.BASE_PACKAGE + ".classfile.*;");
        this.printWriter.println("import " + BCELifier.BASE_PACKAGE + ".*;");
        this.printWriter.println("import java.io.*;");
        this.printWriter.println();
        this.printWriter.println("public class " + className + "Creator {");
        this.printWriter.println("  private InstructionFactory _factory;");
        this.printWriter.println("  private ConstantPoolGen    _cp;");
        this.printWriter.println("  private ClassGen           _cg;");
        this.printWriter.println();
        this.printWriter.println("  public " + className + "Creator() {");
        this.printWriter.println("    _cg = new ClassGen(\"" + (packageName.isEmpty() ? className : (packageName + "." + className)) + "\", \"" + superName + "\", \"" + clazz.getSourceFileName() + "\", " + printFlags(clazz.getAccessFlags(), FLAGS.CLASS) + ", new String[] { " + inter + " });");
        this.printWriter.println("    _cg.setMajor(" + clazz.getMajor() + ");");
        this.printWriter.println("    _cg.setMinor(" + clazz.getMinor() + ");");
        this.printWriter.println();
        this.printWriter.println("    _cp = _cg.getConstantPool();");
        this.printWriter.println("    _factory = new InstructionFactory(_cg, _cp);");
        this.printWriter.println("  }");
        this.printWriter.println();
        this.printCreate();
        final Field[] fields = clazz.getFields();
        if (fields.length > 0) {
            this.printWriter.println("  private void createFields() {");
            this.printWriter.println("    FieldGen field;");
            for (final Field field : fields) {
                field.accept(this);
            }
            this.printWriter.println("  }");
            this.printWriter.println();
        }
        final Method[] methods = clazz.getMethods();
        for (int i = 0; i < methods.length; ++i) {
            this.printWriter.println("  private void createMethod_" + i + "() {");
            methods[i].accept(this);
            this.printWriter.println("  }");
            this.printWriter.println();
        }
        this.printMain();
        this.printWriter.println("}");
    }
    
    @Override
    public void visitMethod(final Method method) {
        final MethodGen mg = new MethodGen(method, this.clazz.getClassName(), this.constantPoolGen);
        this.printWriter.println("    InstructionList il = new InstructionList();");
        this.printWriter.println("    MethodGen method = new MethodGen(" + printFlags(method.getAccessFlags(), FLAGS.METHOD) + ", " + printType(mg.getReturnType()) + ", " + printArgumentTypes(mg.getArgumentTypes()) + ", new String[] { " + Utility.printArray(mg.getArgumentNames(), false, true) + " }, \"" + method.getName() + "\", \"" + this.clazz.getClassName() + "\", il, _cp);");
        this.printWriter.println();
        final BCELFactory factory = new BCELFactory(mg, this.printWriter);
        factory.start();
        this.printWriter.println("    method.setMaxStack();");
        this.printWriter.println("    method.setMaxLocals();");
        this.printWriter.println("    _cg.addMethod(method.getMethod());");
        this.printWriter.println("    il.dispose();");
    }
    
    static {
        BASE_PACKAGE = Const.class.getPackage().getName();
        CONSTANT_PREFIX = Const.class.getSimpleName() + ".";
    }
    
    public enum FLAGS
    {
        UNKNOWN, 
        CLASS, 
        METHOD;
    }
}
